from openfabric_pysdk.fields import Schema, fields, post_load

from openfabric_pysdk.utility import SchemaUtil


#######################################################
#  UID Schema
#######################################################
class UserId(Schema):
    uid: str = None


class UserIdSchema(Schema):
    uid = fields.String(required=True)

    @post_load
    def create(self, data, **kwargs):
        return SchemaUtil.create(UserId(), data)



import logging
from typing import Dict

from openfabric_pysdk.context import Ray, State, StateStatus
from openfabric_pysdk.loader import ConfigClass, InputClass, OutputClass, \
    config_callback_function, \
    execution_callback_function
from openfabric_pysdk.loader.config import manifest, state_config
from openfabric_pysdk.store import KeyValueDB


#######################################################
#  App
#######################################################
class App:
    state: State = None

    # ------------------------------------------------------------------------
    def __init__(self):
        self.state = State()

    # ------------------------------------------------------------------------
    def set_status(self, status: StateStatus):
        self.state.status = status

    # ------------------------------------------------------------------------
    def execution_callback_function(self, input: InputClass, ray: Ray) -> OutputClass:
        return execution_callback_function(input, ray, self.state)

    # ------------------------------------------------------------------------
    def config_callback_function(self, config: Dict[str, ConfigClass]):
        logging.info(f"Openfabric - apply APP configuration")
        if config_callback_function is not None:
            try:
                config_callback_function(config, self.state)
            except Exception as e:
                logging.error(f"Openfabric - invalid configuration can\'t restored : {e}")
        else:
            logging.warning(f"Openfabric - no configuration callback available")

    # ------------------------------------------------------------------------
    def get_manifest(self) -> KeyValueDB:
        return manifest

    # ------------------------------------------------------------------------
    def get_state_config(self) -> KeyValueDB:
        return state_config



import json
import zlib
from typing import Any, Dict

from openfabric_pysdk.benchmark import MeasureBlockTime


#######################################################
#  Zip util
#######################################################
class ZipUtil:

    # ------------------------------------------------------------------------
    @staticmethod
    def decompress(data: bytes) -> Dict[str, Any]:
        with MeasureBlockTime("ZipUtil::decompress"):
            uncompressed = zlib.decompress(data)
            string: str = uncompressed.decode('utf-8')
            return json.loads(string)

    # ------------------------------------------------------------------------
    @staticmethod
    def compress(data: Dict[str, Any]) -> bytes:
        with MeasureBlockTime("ZipUtil::compress"):
            string = json.dumps(data)
            return zlib.compress(string.encode('utf-8'))



from openfabric_pysdk.fields import Schema, fields

from .state import StateStatus


#######################################################
#  State schema
#######################################################
class StateSchema(Schema):
    status = fields.Enum(StateStatus)
    started_at = fields.DateTime()



import logging
import os
import traceback
import uuid
from typing import Any, Dict, List

from openfabric_pysdk.app import App
from openfabric_pysdk.benchmark import MeasureBlockTime
from openfabric_pysdk.context import MessageType, Ray, RaySchema, RayStatus, State
from openfabric_pysdk.loader import InputSchema, OutputSchema
from openfabric_pysdk.store import Store

# TODO: should not exist as it:
#    - creates inconsistent usage ('submitted' message not published, nor periodical 'progress')
#    - adds a way to kill the instance by overloading the system
# Would make more sense to have instead a priority queue and process only one or at most a limited number of requests at a time.
#######################################################
#  Foreground Engine
#######################################################
class ForegroundEngine:
    __app: App = None
    __store: Store = None
    __rays: Dict[str, Ray] = None

    # ------------------------------------------------------------------------
    def __init__(self):
        self.__rays = dict()
        self.__store = Store(path=f"{os.getcwd()}/datastore")

    # ------------------------------------------------------------------------
    def prepare(self, app: App, data: Any, qid=None, sid=None, uid=None, rid=None) -> str:
        if qid is None:
            qid: str = uuid.uuid4().hex
        ray = self.ray(qid)
        ray.status = RayStatus.QUEUED
        ray.sid = sid
        ray.uid = uid
        ray.rid = rid
        self.write(qid, 'ray', ray, RaySchema().dump)
        self.write(qid, 'in', data)
        self.__app = app
        return qid

    # ------------------------------------------------------------------------
    def ray(self, qid: str):
        if self.__rays.get(qid) is None:
            ray = Ray(qid=qid)
            self.__rays[qid] = ray
        return self.__rays[qid]

    # ------------------------------------------------------------------------
    def rays(self, criteria=None) -> List[Ray]:
        rays: List[Ray] = []
        for qid, ray in self.__rays.items():
            if criteria is None or criteria(ray):
                rays.append(ray)
        return rays

    # ------------------------------------------------------------------------
    def process(self, qid):

        with MeasureBlockTime("Engine::execution_callback_function"):

            output = None
            ray = self.ray(qid)
            try:
                data = self.read(qid, 'in', InputSchema().load)
                if data is None:
                    return None

                ray.status = RayStatus.RUNNING
                self.write(qid, 'ray', ray, RaySchema().dump)

                # Callback execution method
                output = self.__app.execution_callback_function(data, ray)

                ray.status = RayStatus.COMPLETED
            except:
                error = f"Openfabric - failed executing: [{qid}]\n{traceback.format_exc()}"
                logging.error(error)
                ray.message(MessageType.ERROR, error)
                ray.status = RayStatus.FAILED
        ray.complete()
        self.write(qid, 'ray', ray, RaySchema().dump)
        self.write(qid, 'out', output, OutputSchema().dump)

        return output

    # ------------------------------------------------------------------------
    def read(self, qid: str, key: str, deserializer=None) -> Any:
        output = self.__store.get(qid, key)
        if output is None:
            return None
        output = deserializer(output) if deserializer is not None else output
        return output

    # ------------------------------------------------------------------------
    def write(self, qid: str, key: str, val: Any, serializer=None):
        if val is not None:
            val = serializer(val) if serializer is not None else val
        self.__store.set(qid, key, val)

    # ------------------------------------------------------------------------
    def delete(self, qid: str) -> Ray:
        ray = self.ray(qid)
        self.__store.drop(qid)
        ray.status = RayStatus.REMOVED
        return ray


foreground = ForegroundEngine()



from typing import Any, Dict

from openfabric_pysdk.flask import *
from openfabric_pysdk.loader import *
from openfabric_pysdk.fields import fields
from openfabric_pysdk.context import RaySchema
from openfabric_pysdk.service import RemoteService
from openfabric_pysdk.transport import ResourceDescriptor


#######################################################
#  Execution Queue API
#######################################################
class QueueGetApi(MethodResource, Resource):
    __descriptor: ResourceDescriptor = None

    # ------------------------------------------------------------------------
    def __init__(self, descriptor: ResourceDescriptor = None):
        self.__descriptor = descriptor

    @doc(description="Get the response for the indicated request", tags=["Queue"])
    @use_kwargs({'qid': fields.String(required=True)}, location='query')
    @marshal_with(OutputSchema)
    def get(self, qid: str, *args) -> OutputClass:
        url = f"{self.__descriptor.remote}/{self.__descriptor.endpoint}?qid={qid}"
        return RemoteService.get(url, deserializer=OutputSchema().load)


class QueueDeleteApi(MethodResource, Resource):
    __descriptor: ResourceDescriptor = None

    # ------------------------------------------------------------------------
    def __init__(self, descriptor: ResourceDescriptor = None):
        self.__descriptor = descriptor

    @doc(description="Remove the indicated requests and the associated results", tags=["Queue"])
    @use_kwargs({'qid': fields.String(required=True)}, location='query')
    @marshal_with(RaySchema)
    def delete(self, qid: str, *args):
        url = f"{self.__descriptor.remote}/{self.__descriptor.endpoint}?qid={qid}"
        return RemoteService.delete(url, deserializer=RaySchema().load)


class QueueListApi(MethodResource, Resource):
    __descriptor: ResourceDescriptor = None

    # ------------------------------------------------------------------------
    def __init__(self, descriptor: ResourceDescriptor = None):
        self.__descriptor = descriptor

    @doc(description="Get list of existing requests and their status", tags=["Queue"])
    @marshal_with(RaySchema(many=True))
    def get(self, *args):
        url = f"{self.__descriptor.remote}/{self.__descriptor.endpoint}"
        return RemoteService.get(url, deserializer=RaySchema(many=True).load)


class QueuePostApi(MethodResource, Resource):
    __descriptor: ResourceDescriptor = None

    # ------------------------------------------------------------------------
    def __init__(self, descriptor: ResourceDescriptor = None):
        self.__descriptor = descriptor

    @doc(description="Queue a new request", tags=["Queue"])
    @use_kwargs(InputSchema, location='json')
    @marshal_with(RaySchema)
    def post(self, *args):
        data = InputSchema().dump(list(args) if InputSchema().many is True else args[0])
        url = f"{self.__descriptor.remote}/{self.__descriptor.endpoint}"
        return RemoteService.post(url, data=data, deserializer=RaySchema().load)



import os
from openfabric_pysdk.store import KeyValueDB

# State configuration
state_config: KeyValueDB = KeyValueDB(f"state", path=f"{os.getcwd()}/config", autodump=True)

# Manifest configuration
manifest: KeyValueDB = KeyValueDB(f"manifest", path=f"{os.getcwd()}/config", autodump=True)

# Execution configuration
execution: KeyValueDB = KeyValueDB(f"execution", path=f"{os.getcwd()}/config", autodump=True)



from hashlib import md5

from openfabric_pysdk.store import LRU


#######################################################
#  Change detector
#######################################################
class ChangeUtil:
    changes = LRU(500)

    @classmethod
    def is_changed(cls, name, content, serializer=None):
        if content is not None:
            content = serializer(content) if serializer is not None else content
        else:
            content = ''
        content_hash = cls.changes.get(name)
        new_content_hash = md5(str(content).encode('utf-8')).hexdigest()
        cls.changes.put(name, new_content_hash)
        return content_hash != new_content_hash



from openfabric_pysdk.fields import Schema, fields


#######################################################
#  Benchmark Schema
#######################################################
class BenchmarkSchema(Schema):
    name = fields.String(allow_none=True)
    avg = fields.String(allow_none=True)
    count = fields.String(allow_none=True)
    stddev = fields.String(allow_none=True)
    min = fields.String(allow_none=True)
    max = fields.String(allow_none=True)

    def __init__(self):
        super().__init__(many=True)



import logging
import os
import traceback

from openfabric_pysdk.flask import *
from openfabric_pysdk.loader import *
from openfabric_pysdk.app import App
from openfabric_pysdk.context import StateStatus
from openfabric_pysdk.execution import Container, Profile


@webserver.route("/")
def index():
    return render_template("index.html", manifest=manifest.all())


class Starter:

    # ------------------------------------------------------------------------
    @staticmethod
    def ignite(debug: bool, host: str, port=5000, remote: str = None):
        # Setup logger
        logger = logging.getLogger()
        if debug is True:
            logger.setLevel(logging.DEBUG)
        else:
            logger.setLevel(logging.INFO)
            logging.getLogger('socketio').setLevel(logging.ERROR)
            logging.getLogger('engineio').setLevel(logging.ERROR)

        # Profile
        profile = Profile()
        profile.host = host
        profile.port = port
        profile.debug = debug
        profile.remote = remote if remote is not None else os.getenv('RAI_URL')

        # Start app execution
        app = App()
        try:
            app.set_status(StateStatus.STARTING)
            container = Container(profile, webserver)
            container.start(app)
        except:
            app.set_status(StateStatus.CRASHED)
            logging.error(f"Openfabric - failed starting app: \n{traceback.format_exc()}")



from typing import List, Union

from openfabric_pysdk.benchmark import MeasureBlockTime
from openfabric_pysdk.flask.rest import *
from openfabric_pysdk.loader import *
from openfabric_pysdk.service import ConfigService
from openfabric_pysdk.transport import ResourceDescriptor
from openfabric_pysdk.transport.schema import UserId, UserIdSchema


#######################################################
#  Config API
#######################################################
class ConfigApi(MethodResource, Resource):
    __descriptor: ResourceDescriptor = None

    # ------------------------------------------------------------------------
    def __init__(self, descriptor: ResourceDescriptor = None):
        self.__descriptor = descriptor

    # ------------------------------------------------------------------------
    @doc(description="Get APP configuration", tags=["Developer"])
    @use_kwargs(UserIdSchema, location='query')
    @marshal_with(ConfigSchema)
    def get(self, uid: UserId) -> ConfigClass:
        with MeasureBlockTime("ConfigRestApi::get"):
            app = self.__descriptor.app
            return ConfigService.read(app, uid)

    # ------------------------------------------------------------------------
    @doc(description="Set APP configuration", tags=["Developer"])
    @use_kwargs(UserIdSchema, location='query')
    @use_kwargs(ConfigSchema, location='json')
    @marshal_with(ConfigSchema)
    def post(self, uid: UserId, config: Union[ConfigClass, List[ConfigClass]]) -> ConfigClass:
        with MeasureBlockTime("ConfigRestApi::post"):
            app = self.__descriptor.app
            return ConfigService.write(app, uid, config)



from openfabric_pysdk.flask import *
from openfabric_pysdk.loader import *
from openfabric_pysdk.service import RemoteService
from openfabric_pysdk.transport import ResourceDescriptor


#######################################################
#  Execution API
#######################################################
class ExecutionApi(MethodResource, Resource):
    __descriptor: ResourceDescriptor = None

    # ------------------------------------------------------------------------
    def __init__(self, descriptor: ResourceDescriptor = None):
        self.__descriptor = descriptor

    @doc(description="Execute execution and get response", tags=["Execution"])
    @use_kwargs(InputSchema, location='json')
    @marshal_with(OutputSchema)
    def post(self, *args) -> OutputClass:
        data = InputSchema().dump(list(args) if InputSchema().many is True else args[0])
        url = f"{self.__descriptor.remote}/{self.__descriptor.endpoint}"
        return RemoteService.post(url, data, OutputSchema().load)



import logging
import threading
import traceback
from time import sleep
from typing import Any, List

from openfabric_pysdk.context import Ray, RaySchema
from openfabric_pysdk.task import Task
from openfabric_pysdk.app import App
from .foreground_engine import foreground


#######################################################
#  Background Engine
#######################################################
class BackgroundEngine:
    __task: Task = None
    __instances: int = 0
    __running: bool = False
    __current_qid: str = None
    __worker: threading.Thread = None
    __lock: threading.Condition = threading.Condition()

    # ------------------------------------------------------------------------
    def __init__(self):
        self.__lock.acquire()
        if self.__instances == 0:
            self.__task = Task()
            self.__worker = threading.Thread(target=self.__process, args=())
            self.__worker.start()

        self.__instances = self.__instances + 1
        self.__lock.release()

        # Wait for processing thread to start
        while not self.__running:
            sleep(0.1)

    # ------------------------------------------------------------------------
    def __del__(self):
        self.__lock.acquire()
        if self.__instances > 0:
            self.__lock.release()
            return

        self.__running = False

        self.__lock.notify_all()
        self.__lock.release()

    # ------------------------------------------------------------------------
    def __process(self):
        self.__running = True
        while self.__running:
            self.__lock.acquire()
            self.__current_qid = None
            while self.__running and self.__task.empty():
                self.__lock.wait()
            try:
                self.__current_qid = self.__task.next()
            except:
                logging.warning("Openfabric - queue empty!")
                traceback.print_exc()
            finally:
                self.__lock.release()

            if self.__running and self.__current_qid is not None:
                foreground.process(self.__current_qid)

    # ------------------------------------------------------------------------
    def prepare(self, app: App, data: str, qid=None, sid=None, uid=None, rid=None) -> str:
        self.__lock.acquire()
        qid: str = foreground.prepare(app, data, qid=qid, sid=sid, uid=uid, rid=rid)
        self.__task.add(qid)
        self.__lock.notify_all()
        self.__lock.release()
        return qid

    # ------------------------------------------------------------------------
    def ray(self, qid: str) -> Ray:
        return foreground.ray(qid)

    # ------------------------------------------------------------------------
    def rays(self, criteria=None) -> List[str]:
        rays: List[str] = []
        for qid in self.__task.all():
            ray = foreground.read(qid, 'ray', RaySchema().load)
            if ray is None:
                continue
            if criteria is None or criteria(ray):
                rays.append(ray)
        rays.sort(key=lambda r: r.created_at)
        return rays

    # ------------------------------------------------------------------------
    def delete(self, qid: str) -> Ray:
        self.__lock.acquire()
        self.__task.rem(qid)
        ray = foreground.delete(qid)
        self.__lock.notify_all()
        self.__lock.release()
        return ray

    # ------------------------------------------------------------------------
    def read(self, qid: str, key: str, deserializer=None) -> Any:
        return foreground.read(qid, key, deserializer)

    # ------------------------------------------------------------------------
    def write(self, qid: str, key: str, val: Any, serializer=None):
        return foreground.write(qid, key, val, serializer)


background = BackgroundEngine()



from typing import Any, Dict

#######################################################
#  Schema util
#######################################################
class SchemaUtil:

    @staticmethod
    def create(instance: Any, data: Dict[Any, Any]):
        for key in data:
            instance.__setattr__(key, data[key])
        return instance



import logging

from openfabric_pysdk.flask import WebserverRestDoc, WebserverRestAPI
from openfabric_pysdk.transport import ResourceDescriptor


#######################################################
#  Rest service
#######################################################
class RestService:

    # ------------------------------------------------------------------------
    @staticmethod
    def install(descriptor: ResourceDescriptor, rest: WebserverRestAPI = None, docs: WebserverRestDoc = None):

        kwargs = dict(descriptor=descriptor)

        if rest is not None:
            logging.info(f"Openfabric - install {descriptor.handler} REST endpoints on {descriptor.endpoint}")
            rest.add_resource(descriptor.handler, descriptor.endpoint, resource_class_kwargs=kwargs)

        if docs is not None:
            logging.info(f"Openfabric - install {descriptor.handler} DOCS endpoints on {descriptor.endpoint}")
            docs.register(descriptor.handler, resource_class_kwargs=kwargs)



from .config import manifest, execution, state_config
from openfabric_pysdk.utility import LoaderUtil

# Output concept
InputClass = LoaderUtil.get_class("input_class")
InputSchema = LoaderUtil.get_class("input_schema")

# Output concept
OutputClass = LoaderUtil.get_class("output_class")
OutputSchema = LoaderUtil.get_class("output_schema")

# Config concept
ConfigClass = LoaderUtil.get_class("config_class")
ConfigSchema = LoaderUtil.get_class("config_schema")

# Execution callback function
execution_callback_function = LoaderUtil.get_function("main_callback")

# Config callback function
config_callback_function = LoaderUtil.get_function("config_callback")



import uuid

from openfabric_pysdk.context import RaySchema
from openfabric_pysdk.engine import background
from openfabric_pysdk.fields import fields
from openfabric_pysdk.flask.rest import *
from openfabric_pysdk.loader import *
from openfabric_pysdk.transport import ResourceDescriptor


#######################################################
#  Execution Queue API
#######################################################
class QueueGetApi(MethodResource, Resource):
    __descriptor: ResourceDescriptor = None

    # ------------------------------------------------------------------------
    def __init__(self, descriptor: ResourceDescriptor = None):
        self.__descriptor = descriptor

    @doc(description="Get the response for the indicated request", tags=["Queue"])
    @use_kwargs({'qid': fields.String(required=True)}, location='query')
    @marshal_with(OutputSchema)
    def get(self, qid: str, *args) -> OutputClass:
        return background.read(qid, 'out', OutputSchema().load)


class QueueDeleteApi(MethodResource, Resource):
    __descriptor: ResourceDescriptor = None

    # ------------------------------------------------------------------------
    def __init__(self, descriptor: ResourceDescriptor = None):
        self.__descriptor = descriptor

    @doc(description="Remove the indicated requests and the associated results", tags=["Queue"])
    @use_kwargs({'qid': fields.String(required=True)}, location='query')
    @marshal_with(RaySchema)
    def delete(self, qid: str, *args):
        return background.delete(qid)


class QueueListApi(MethodResource, Resource):
    __descriptor: ResourceDescriptor = None

    # ------------------------------------------------------------------------
    def __init__(self, descriptor: ResourceDescriptor = None):
        self.__descriptor = descriptor

    @doc(description="Get list of existing requests and their status", tags=["Queue"])
    @marshal_with(RaySchema(many=True))
    def get(self, *args):
        return background.rays()


class QueuePostApi(MethodResource, Resource):
    __descriptor: ResourceDescriptor = None

    # ------------------------------------------------------------------------
    def __init__(self, descriptor: ResourceDescriptor = None):
        self.__descriptor = descriptor

    @doc(description="Queue a new request", tags=["Queue"])
    @use_kwargs(InputSchema, location='json')
    @marshal_with(RaySchema)
    def post(self, *args):
        data = InputSchema().dump(list(args) if InputSchema().many is True else args[0])
        sid = uuid.uuid4().hex
        uid = uuid.uuid4().hex
        rid = uuid.uuid4().hex
        app = self.__descriptor.app
        qid = background.prepare(app, data, sid=sid, uid=uid, rid=rid)
        ray = background.ray(qid)
        return ray



from openfabric_pysdk.app import App
from openfabric_pysdk.context import State


#######################################################
#  Resource descriptor
#######################################################
class ResourceDescriptor:
    app: App
    remote: str
    handler: type
    endpoint: str



from openfabric_pysdk.utility import SchemaUtil
from openfabric_pysdk.fields import Schema, fields, post_load

from .ray import Ray, RayStatus
from .bar_schema import BarSchema
from .message_schema import MessageSchema


#######################################################
#  Ray schema
#######################################################
class RaySchema(Schema):
    sid = fields.String()
    uid = fields.String()
    qid = fields.String()
    rid = fields.String()
    bars = fields.Dict(
        keys=fields.Str(),
        values=fields.Nested(BarSchema),
        allow_none=True
    )
    messages = fields.Nested(MessageSchema(many=True))
    status = fields.Enum(RayStatus)
    created_at = fields.DateTime()
    updated_at = fields.DateTime()

    @post_load
    def create(self, data, **kwargs):
        return SchemaUtil.create(Ray(None), data)



import gc
import logging
import timeit
from functools import wraps
from typing import Dict, Optional

from runstats import Statistics

_stats: Dict[str, Statistics] = {}
_lasts: Dict[str, float] = {}


def add_timing(name: str, elapsed: float, no_print=True) -> Statistics:
    global _stats
    global _lasts

    stats = _stats.get(name, None)
    if stats is None:
        stats = Statistics()
        _stats[name] = stats
    stats.push(elapsed)
    _lasts[name] = elapsed

    if not no_print:
        logging.debug('Openfabric - timing "{}": {}s'.format(name, elapsed))
    return stats


def get_last(name: str) -> float:
    global _lasts
    return _lasts[name]


def get_timing(name: str) -> Statistics:
    global _stats
    return _stats.get(name)


def get_all_timings() -> Dict[str, Statistics]:
    global _stats
    return _stats


def get_all_timings_json():
    global _stats
    res = []
    for name in _stats:
        stats = _stats[name]
        count = len(stats)
        res.append({
            "name": f'{name}',
            'avg': f'{stats.mean():.2f}',
            'count': f'{count}',
            'stddev': f'{stats.stddev() if count > 1 else float("NaN"):.2f}',
            'min': f'{stats.minimum():.2f}',
            'max': f'{stats.maximum():.2f}',
        })

    return res


def print_all_timing() -> None:
    global _stats
    for name in _stats:
        print_timing(name)


def print_timing(name: str) -> None:
    global _stats
    stats = _stats.get(name, None)
    if stats is None:
        logging.debug(f'Openfabric - timing_name="{name}", avg=never_recorded')
    else:
        count = len(stats)
        logging.debug(f'Openfabric - timing_name="{name}", '
                      f'avg={stats.mean():.4g} '
                      f'count={count} '
                      f'stddev={stats.stddev() if count > 1 else float("NaN"):.4g} '
                      f'min={stats.minimum():.4g} '
                      f'max={stats.maximum():.4g} '
                      )


def clear_timings() -> None:
    global _stats
    _stats.clear()


def MeasureTime(f_py=None, no_print=True, disable_gc=False, name: Optional[str] = None):
    assert callable(f_py) or f_py is None

    def _decorator(f):
        @wraps(f)
        def _wrapper(*args, **kwargs):
            gcold = gc.isenabled()
            if disable_gc:
                gc.disable()
            start_time = timeit.default_timer()
            try:
                result = f(*args, **kwargs)
            finally:
                elapsed = timeit.default_timer() - start_time
                if disable_gc and gcold:
                    gc.enable()
                fname = name or f.__name__
                add_timing(fname, elapsed, no_print=no_print)
            return result

        return _wrapper

    return _decorator(f_py) if callable(f_py) else _decorator


class MeasureBlockTime:
    def __init__(self, name: str, no_print=False, disable_gc=False):
        self.name = name
        self.no_print = no_print
        self.disable_gc = disable_gc

    def cur_elapsed(self) -> float:
        return timeit.default_timer() - self.start_time

    def __enter__(self):
        self.gcold = gc.isenabled()
        if self.disable_gc:
            gc.disable()
        self.start_time = timeit.default_timer()
        return self

    def __exit__(self, ty, val, tb):
        self.elapsed = self.cur_elapsed()
        if self.disable_gc and self.gcold:
            gc.enable()
        self.stats = add_timing(self.name, self.elapsed, no_print=self.no_print)
        return False  # re-raise any exceptions



from typing import Dict, List, Union

from openfabric_pysdk.app import App
from openfabric_pysdk.loader import ConfigClass, ConfigSchema
from openfabric_pysdk.transport.schema import UserId


#######################################################
#  Config service
#######################################################
class ConfigService:

    # ------------------------------------------------------------------------
    @staticmethod
    def read(app: App, uidc: UserId) -> ConfigClass:
        state_config = app.get_state_config()
        config = state_config.get(uidc.uid)
        if config is False:
            return dict()
        return config

    # ------------------------------------------------------------------------
    @staticmethod
    def write(app: App, uidc: UserId, config: Union[ConfigClass, List[ConfigClass]]) -> Dict[str, ConfigClass]:
        if ConfigSchema().many is True and not isinstance(config, list):
            config = [config]
        else:
            config = config[0] if type(config) == list else config
        # Store
        state_config = app.get_state_config()
        state_config.set(uidc.uid, ConfigSchema().dump(config))
        # Apply configuration
        ConfigService.apply(app)

        return config

    # ------------------------------------------------------------------------
    @staticmethod
    def apply(app: App):
        all_items = ConfigService.read_all(app)
        app.config_callback_function(all_items)

    # ------------------------------------------------------------------------
    @staticmethod
    def read_all(app: App) -> Dict[str, ConfigClass]:
        # Reload form config file
        state_config = app.get_state_config()
        items = state_config.all().items()
        return dict(map(lambda kv: (kv[0], ConfigSchema().load(kv[1])), items))



from .rest_service import RestService
from .socket_service import SocketService
from .swagger_service import SwaggerService
from .remote_service import RemoteService
from .config_service import ConfigService



from openfabric_pysdk.utility import SchemaUtil
from openfabric_pysdk.fields import Schema, fields, post_load

from .message import Message, MessageType


#######################################################
#  Message schema
#######################################################

class MessageSchema(Schema):
    type = fields.Enum(MessageType)
    content = fields.String()
    created_at = fields.DateTime()

    # ------------------------------------------------------------------------
    @post_load
    def create(self, data, **kwargs):
        return SchemaUtil.create(Message(), data)



import logging
from typing import List

from openfabric_pysdk.flask.core import Webserver
from openfabric_pysdk.flask.socket import SocketIOClient, SocketIOServer
from openfabric_pysdk.transport import ResourceDescriptor


#######################################################
#  Socket service
#######################################################
class SocketService:

    # ------------------------------------------------------------------------
    @staticmethod
    def install(descriptor: ResourceDescriptor, webserver: Webserver):
        logging.info(f"Openfabric - install Execution SOCKET endpoints on {descriptor.endpoint}")
        return descriptor.handler(webserver, descriptor)

    # ------------------------------------------------------------------------
    @staticmethod
    def server(webserver: Webserver):
        # Set this variable to "threading", "eventlet" or "gevent" to test the
        # different async modes, or leave it set to None for the application to choose
        # the best option based on installed packages.
        async_mode = "gevent"
        max_size = 10000000 * 100  # 100Mb
        return SocketIOServer(
            webserver=webserver,
            async_mode=async_mode,
            cors_allowed_origins='*',
            max_http_buffer_size=max_size
        )

    # ------------------------------------------------------------------------
    @staticmethod
    def client(url: str, namespaces: List[str]):
        client = SocketIOClient()
        client.connect(url, namespaces=namespaces)
        return client

    # ------------------------------------------------------------------------
    @staticmethod
    def create_portal(source: [SocketIOServer, SocketIOClient], target: [SocketIOServer, SocketIOClient],
                      namespace: str, webserver: Webserver, events: List[str]):
        app_context = webserver.app_context()

        def _bind_handler(event_name):
            @source.on(event_name, namespace=namespace)
            def _handler(*args):
                with app_context:
                    target.emit(event_name, *args, namespace=namespace)

        for event in events:
            _bind_handler(event)



import logging
from typing import Any, Set
from datetime import datetime

import gevent
import uuid
import copy

from openfabric_pysdk.app import App
from openfabric_pysdk.benchmark import MeasureBlockTime
from openfabric_pysdk.context import Ray, RaySchema, State, StateSchema
from openfabric_pysdk.engine import background, foreground
from openfabric_pysdk.flask.core import Webserver, request
from openfabric_pysdk.flask.socket import Namespace, SocketIOServer, emit
from openfabric_pysdk.service import SocketService
from openfabric_pysdk.transport import ResourceDescriptor
from openfabric_pysdk.transport.schema import UserIdSchema
from openfabric_pysdk.utility import ChangeUtil, ZipUtil
from openfabric_pysdk.service import ConfigService


#######################################################
#  Execution Socket
#######################################################
class ExecutionSocket(Namespace):
    __app: App = None
    __sessions: Set[str] = None
    __webserver: Webserver = None
    __sioserver: SocketIOServer = None

    # ------------------------------------------------------------------------
    def __init__(self, webserver: Webserver, descriptor: ResourceDescriptor):
        super().__init__(descriptor.endpoint)
        self.__sessions = set()
        self.__app = descriptor.app
        self.__webserver = webserver

    # ------------------------------------------------------------------------
    def run(self, debug: bool, host: str, port: int):
        self.__sioserver = SocketService.server(self.__webserver)
        self.__sioserver.on_namespace(self)
        self.__sioserver.run(host=host, debug=debug, port=port)

    # ------------------------------------------------------------------------
    def on_configure(self, config: bytes):
        with MeasureBlockTime("Socket::configure"):
            sid = request.sid

            dictionary = ZipUtil.decompress(config)
            config = dictionary.get('body', None)
            header = dictionary.get('header', dict())
            uid = header.get('uid', None)
            user = UserIdSchema().load(dict(uid=uid))

            if config != None:
                ConfigService.write(self.__app, user, config)
            else:
                # Return existing configuration
                # TODO: or should we have the option to clear it?
                config = ConfigService.read(self.__app, user)

            emit('settings', dict(uid=uid, config=config))

    # ------------------------------------------------------------------------
    def on_execute(self, data: bytes, background: bool):
        with MeasureBlockTime("Socket::execute"):
            sid = request.sid

            dictionary = ZipUtil.decompress(data)
            data = dictionary.get('body', None)
            header = dictionary.get('header', dict())
            uid = header.get('uid', None)
            rid = header.get('rid', uuid.uuid4().hex)

            logging.info(f"Socket::execute: sid={sid}, uid={uid}, rid={rid}, background={background}")
            if background is True:
                self.__execute_background(data, sid, uid, rid)
            else:
                self.__execute_foreground(data, sid, uid, rid)

    def __execute_foreground(self, data: Any, sid: str, uid: str, rid: str):
        # Setup
        qid = foreground.prepare(self.__app, data, qid=sid, sid=sid, uid=uid, rid=rid)
        ray = foreground.ray(qid)
        # Execute in foreground
        with MeasureBlockTime("Socket::callback"):
            foreground.process(qid)
            output = foreground.read(qid, 'out')
            ray_dump = RaySchema().dump(ray)
            foreground.delete(qid)
            emit('response', dict(output=output, ray=ray_dump))

    def __execute_background(self, data: str, sid: str, uid: str, rid: str):
        maximum_update_interval = 3 # seconds
        # Setup
        qid = background.prepare(self.__app, data, sid=sid, uid=uid, rid=rid)
        ray = background.ray(qid)
        prevRay = None
        emit('submitted', RaySchema().dump(ray))

        last_update = now = datetime.now()

        # Execute in background
        with MeasureBlockTime("Socket::callback"):
            while sid in self.__sessions:
                if prevRay == None or prevRay.updated_at != ray.updated_at:
                    try:
                        prevRay = copy.deepcopy(ray)

                        ray_dump = RaySchema().dump(ray)
                        emit('progress', ray_dump)
                        last_update = datetime.now()
                    except:
                        # Failed to copy ray. Downside is that same message will be sent again.
                        pass

                if (datetime.now() - last_update).total_seconds() > maximum_update_interval:
                    last_update = datetime.now()
                    ray_dump = RaySchema().dump(ray)
                    emit('progress', ray_dump)

                if ray.finished is True:
                    ray_dump = RaySchema().dump(ray)
                    # input = background.get(qid, 'in')
                    output = background.read(qid, 'out')
                    emit('response', dict(output=output, ray=ray_dump))
                    break

                gevent.sleep(0.1)

    # ------------------------------------------------------------------------
    def on_resume(self, uid: str):
        sid = request.sid
        with MeasureBlockTime("OpenfabricSocket::restore"):
            def criteria(_ray: Ray):
                # is deleted ?
                if _ray is None:
                    return False
                # is different user ?
                if _ray.uid != uid:
                    return False
                # is an active session available ?
                if _ray.sid in self.__sessions and _ray.sid != sid:
                    return False
                return True

            # Filter rays
            rays = background.rays(criteria)
            for ray in rays:
                emit('progress', RaySchema().dump(ray))

    # ------------------------------------------------------------------------
    def on_restore(self, qid: str):
        ray_dump = foreground.read(qid, 'ray')
        input_dump = foreground.read(qid, 'in')
        output_dump = foreground.read(qid, 'out')
        emit('restore', dict(input=input_dump, output=output_dump, ray=ray_dump))

    # ------------------------------------------------------------------------
    def on_state(self, uid: str):
        sid = request.sid
        with MeasureBlockTime("Socket::state"):
            while sid in self.__sessions:
                changed = ChangeUtil.is_changed('execution::state' + sid, self.__app.state, StateSchema().dump)
                if changed is True:
                    emit('state', StateSchema().dump(self.__app.state))
                gevent.sleep(0.1)

    # ------------------------------------------------------------------------
    def on_delete(self, qid: str):
        sid = request.sid
        with MeasureBlockTime("OpenfabricSocket::delete"):
            foreground.delete(qid)
            background.delete(qid)

    # ------------------------------------------------------------------------
    def on_connect(self):
        sid = request.sid
        logging.debug(f'Openfabric - client connected {sid} on {request.host}')
        self.__sessions.add(sid)

    # ------------------------------------------------------------------------
    def on_disconnect(self):
        sid = request.sid
        logging.debug(f'Openfabric - client disconnected {sid} on {request.host}')
        self.__sessions.remove(sid)



import importlib
from pydoc import locate

from openfabric_pysdk.loader.config import execution
from openfabric_pysdk.benchmark import MeasureBlockTime


#######################################################
# Class loader utility class
#######################################################
class LoaderUtil:

    @staticmethod
    def import_class(name):
        return locate(name.__module__ + "." + name.__name__)

    @staticmethod
    def get_class(name):
        with MeasureBlockTime("Loader::get_class"):
            class_config = execution.get(name)
            if class_config is None:
                return None
            clazz = class_config['class']
            package = class_config['package']
            if clazz is None or package is None:
                return None
            module = importlib.import_module(package)
            return locate(module.__name__ + "." + clazz)

    @staticmethod
    def get_function(name):
        with MeasureBlockTime("Loader::get_function"):
            function_config = execution.get(name)
            if function_config is None:
                return None
            function = function_config['function']
            package = function_config['package']
            if function is None or package is None:
                return None
            module = importlib.import_module(package)
            # print(module.__name__ + "." + function)
            return locate(module.__name__ + "." + function)



from openfabric_pysdk.benchmark import MeasureBlockTime, get_all_timings_json
from openfabric_pysdk.flask import *
from openfabric_pysdk.transport import ResourceDescriptor
from openfabric_pysdk.transport.schema import BenchmarkSchema


#######################################################
#  Benchmark API
#######################################################
class BenchmarkApi(MethodResource, Resource):
    __descriptor: ResourceDescriptor = None

    # ------------------------------------------------------------------------
    def __init__(self, descriptor: ResourceDescriptor = None):
        self.__descriptor = descriptor

    # ------------------------------------------------------------------------
    @doc(description="Get APP benchmarks", tags=["Developer"])
    @marshal_with(BenchmarkSchema)
    def get(self):
        with MeasureBlockTime("BenchmarkRestApi::get"):
            return get_all_timings_json()



import json
from typing import Any

from openfabric_pysdk.flask import *


#######################################################
#  Remote service
#######################################################
class RemoteService:

    # ------------------------------------------------------------------------
    @staticmethod
    def get(url: str, deserializer=None):
        return RemoteService.__unwrap(get(url), deserializer=deserializer)

    # ------------------------------------------------------------------------
    @staticmethod
    def post(url: str, data: Any, deserializer=None):
        return RemoteService.__unwrap(post(url, json=data), deserializer=deserializer)

    # ------------------------------------------------------------------------
    @staticmethod
    def patch(url: str, data: Any, deserializer=None):
        return RemoteService.__unwrap(patch(url, json=data), deserializer=deserializer)

    # ------------------------------------------------------------------------
    @staticmethod
    def delete(url: str, deserializer=None):
        return RemoteService.__unwrap(delete(url), deserializer=deserializer)

    # ------------------------------------------------------------------------
    @staticmethod
    def __unwrap(response: Response, deserializer=None):
        if response.ok is False:
            return None
        json_object = json.loads(response.text)
        return deserializer(json_object)



import logging

from openfabric_pysdk.flask import SwaggerMarshmallowPlugin, Webserver, WebserverSwaggerSpec
from openfabric_pysdk.transport import ResourceDescriptor


#######################################################
#  Swagger service
#######################################################
class SwaggerService:

    # ------------------------------------------------------------------------
    @staticmethod
    def install(descriptor: ResourceDescriptor, webserver: Webserver):
        logging.info(f"Openfabric - install Specs REST endpoints on {descriptor.endpoint}")
        manifest = descriptor.app.get_manifest()
        specs = {
            'APISPEC_SPEC': WebserverSwaggerSpec(
                title="App " + manifest.get('name'),
                version=manifest.get('version'),
                plugins=[SwaggerMarshmallowPlugin()],
                openapi_version='2.0.0',
                info=dict(
                    termsOfService='https://openfabric.ai/terms/',
                    contact=dict(
                        name=manifest.get('organization'),
                        url="https://openfabric.ai"
                    ),
                    description=manifest.get('description')),
            ),
            'APISPEC_SWAGGER_URL': f'/{descriptor.endpoint}/',  # URI to access API Doc JSON
            'APISPEC_SWAGGER_UI_URL': f'/{descriptor.endpoint}-ui/'  # URI to access UI of API Doc
        }
        webserver.config.update(specs)



from openfabric_pysdk.utility import SchemaUtil
from openfabric_pysdk.fields import Schema, fields, post_load

from .bar import Bar


#######################################################
#  Bar schema
#######################################################
class BarSchema(Schema):
    percent = fields.String(allow_none=True)
    remaining = fields.String(allow_none=True)

    @post_load
    def create(self, data, **kwargs):
        return SchemaUtil.create(Bar(), data)



from datetime import datetime
from enum import Enum
from typing import Dict, Set

from tqdm.asyncio import tqdm, tqdm_asyncio

from .bar import Bar
from .message import Message, MessageType


#######################################################
#  Ray Status
#######################################################
class RayStatus(Enum):
    QUEUED = 'queued',
    PENDING = 'pending',
    COMPLETED = 'completed',
    RUNNING = 'running',
    CANCELED = 'canceled',
    REMOVED = 'removed',
    UNKNOWN = 'unknown',
    FAILED = 'failed'

    # ------------------------------------------------------------------------
    def __str__(self):
        return self.name


#######################################################
#  Ray
#######################################################
class Ray:
    uid: str = None
    sid: str = None
    qid: str = None
    rid: str = None
    finished: bool = None
    bars: Dict[str, Bar] = None
    status: RayStatus = None
    created_at: datetime = None
    updated_at: datetime = None
    messages: Set[Message] = None
    __tqdms: Dict[str, tqdm_asyncio] = None

    # ------------------------------------------------------------------------
    def __init__(self, qid):
        self.__tqdms = dict()
        self.qid = qid
        self.bars = dict(default=Bar())
        self.finished = False
        self.status = RayStatus.UNKNOWN
        self.created_at = datetime.now()
        self.updated_at = datetime.now()
        self.messages = set()

    # ------------------------------------------------------------------------
    def update(self, other):
        for name, value in vars(other).items():
            if name in ["qid", "sid", "uid", "rid"]:
                continue
            setattr(self, name, value)

    # ------------------------------------------------------------------------
    def complete(self, name='default'):
        self.finished = True
        bar = self.bars.get(name, Bar())
        self.bars[name] = bar
        bar.remaining = 0
        bar.percent = 100

    # ------------------------------------------------------------------------
    def progress(self, name='default', step=1, total=100) -> tqdm_asyncio:
        if self.__tqdms.get(name) is None:
            self.__tqdms[name] = tqdm(total=total)
        tqdm_bar = self.__tqdms[name]
        # --
        bar = self.bars.get(name, Bar())
        self.bars[name] = bar
        f_dict = tqdm_bar.format_dict
        rate = f_dict.get("rate")
        total = tqdm_bar.total
        n = tqdm_bar.n
        remaining = (total - n) / rate if rate and total else 0
        bar.remaining = max(0, remaining)
        bar.percent = n
        # --
        tqdm_bar.update(step)
        return tqdm_bar

    # ------------------------------------------------------------------------
    def message(self, message_type: MessageType, content: str):
        message = Message()
        message.type = message_type
        message.content = content
        self.messages.add(message)

    # ------------------------------------------------------------------------
    def clear_messages(self):
        self.messages.clear()

    # ------------------------------------------------------------------------
    def tqdms(self):
        return self.__tqdms



from openfabric_pysdk.fields import Schema, fields


#######################################################
#  Manifest Schema
#######################################################
class ManifestSchema(Schema):
    name = fields.String(description="App name", allow_none=True)
    version = fields.String(description="App version", allow_none=True)
    description = fields.String(description="APP description", allow_none=True)
    organization = fields.String(description="APP organization", allow_none=True)
    sdk = fields.String(description="APP sdk", allow_none=True)
    overview = fields.String(description="APP overview", allow_none=True)
    input = fields.String(description="APP input", allow_none=True)
    output = fields.String(description="APP output", allow_none=True)

    def __init__(self):
        super().__init__(many=False)



from .bar import Bar
from .ray import Ray, RayStatus
from .state import State, StateStatus
from .message import Message, MessageType

from .bar_schema import BarSchema
from .ray_schema import RaySchema
from .message_schema import MessageSchema
from .state_schema import StateSchema



# TODO: should not exist
from .config_api import ConfigApi
from .manifest_api import ManifestApi
from .benchmark_api import BenchmarkApi
from .execution_api import ExecutionApi
from .queue_api import QueueGetApi, QueueDeleteApi, QueueListApi, QueuePostApi



import logging
import os
import queue
from enum import Enum
from typing import List, Any

from openfabric_pysdk.store import Store

STATE = 'state'


#######################################################
#  TaskType
#######################################################
class TaskType(Enum):
    QUEUED = 'queued',
    REQUESTED = 'requested',
    COMPLETED = 'completed',

    def __str__(self):
        return self.name


#######################################################
#  Task
#######################################################
class Task:
    __tasks: queue.Queue = None
    __store: Store = None

    # ------------------------------------------------------------------------
    def __init__(self):
        self.__tasks = queue.Queue()
        self.__store = Store(path=f"{os.getcwd()}/datastore", autodump=True)
        pending_tasks = self.__load(TaskType.QUEUED)
        if pending_tasks:
            logging.info(f"Openfabric - restore pre-existing tasks: {pending_tasks}")
            [self.__tasks.put(sid) for sid in pending_tasks]

    # ------------------------------------------------------------------------
    def __del__(self):
        self.__tasks = queue.Queue()

    # ------------------------------------------------------------------------
    def empty(self) -> bool:
        return self.__tasks.empty()

    # ------------------------------------------------------------------------
    def next(self) -> str:
        tid = self.__tasks.get(False)
        self.__save(TaskType.QUEUED, list(self.__tasks.queue))
        self.__append(TaskType.COMPLETED, tid)
        return tid

    # ------------------------------------------------------------------------
    def add(self, tid: str):
        self.__tasks.put(tid)
        self.__save(TaskType.QUEUED, list(self.__tasks.queue))
        self.__append(TaskType.REQUESTED, tid)

    # ------------------------------------------------------------------------
    def rem(self, tid: str):
        self.__remove(TaskType.REQUESTED, tid)
        self.__remove(TaskType.COMPLETED, tid)

    # ------------------------------------------------------------------------
    def all(self):
        return list(
            set(
                self.__load(TaskType.REQUESTED, list()) +
                self.__load(TaskType.REQUESTED, list()) +
                self.__load(TaskType.QUEUED, list())
            )
        )

    # ------------------------------------------------------------------------
    def __append(self, task_type: TaskType, tid: str):
        entries: List[str] = self.__load(task_type, list())
        entries.append(tid)
        self.__save(task_type, entries)

    def __remove(self, task_type: TaskType, tid: str):
        entries: List[str] = self.__load(task_type, list())
        if tid in entries:
            entries.remove(tid)
        self.__save(task_type, entries)

    def __save(self, task_type: TaskType, value: Any):
        self.__store.set(STATE, str(task_type), value)

    def __load(self, task_type: TaskType, default=None):
        return self.__store.get(STATE, str(task_type), default)



# ----------------------------
# !!!  Perform GEVENT monkey patch
# ----------------------------
from gevent import monkey
monkey.patch_all()

import pathlib
from flask import Flask as Webserver, render_template, request, session
from flask_cors import CORS
from requests import get, post, patch, delete, Request, Response

webserver = Webserver(__name__, template_folder=f"{pathlib.Path(__file__).parent}/../../view")
CORS(webserver)



#!/usr/bin/env python3

import json
import logging
import threading
import collections

from typing import Any, Dict

from datetime import datetime
import threading
import socketio, time
from concurrent.futures import ThreadPoolExecutor

import uuid

import zlib

# forward declaration
class Proxy: pass

class ExecutionResult:
    __rid = None
    __proxy = None
    __result = None
    __progress = None
    __finished = None
    __last_update = now = datetime.now()

    def __init__(self, proxy: Proxy):
        self.__proxy = proxy
        self.__rid = uuid.uuid4().hex
        self.__finished = False
        logging.info(f"Created rid" + str(self.__rid))

    def __del__(self):
        self.cancel()

    def on_progress(self, data):
        self.__progress = data

    def on_response(self, data):
        if data != None and "ray" in data:
            self.__progress = data["ray"]

        if data != None and "output" in data:
            self.__result = data["output"]

        self.__finished = True
        self.cancel()

    def on_restore(self, data):
        if data != None and "ray" in data:
            self.__progress = data["ray"]

        if data != None and "output" in data:
            # TODO: if output != None: force finish?
            self.__result = data["output"]

    def request_id(self):
        return self.__rid

    def messages(self):
        if self.__progress != None and "messages" in self.__progress:
            return self.__progress["messages"]

        return []

    def progress(self, name="default"):
        if self.__progress != None and "bars" in self.__progress:
            if name in self.__progress["bars"]:
                return self.__progress["bars"][name]

        return None

    def status(self):
        if self.__progress != None and "status" in self.__progress:
            return self.__progress["status"]

        return "UNKNOWN"

    def cancel(self):
        if self.__proxy != None:
            self.__proxy.cancel(self)
            self.__proxy = None

        if self.__finished == False:
            if self.__progress == None:
                self.__progress = {}
            self.__progress["status"] = "CANCELLED"

    def wait(self, timeout=0):
        remaining = timeout

        while self.__proxy != None and (timeout == 0 or remaining > 0) and self.__finished == False:
            remaining -= 0.1
            time.sleep(0.1)

        return self.__finished

    def data(self):
        return self.__result

    def __repr__(self):
        return f"ExecutionResult({self.__rid}, {self.__progress}, {self.__result})"

    def __str__(self):
        return self.__repr__()

#######################################################
#  Proxy
#######################################################
class Proxy:
    __sio = socketio.Client()
    __url = None
    __tag = None
    __cancel_next = False
    __pending_responses = 0
    __runner: threading.Thread = None
    __checker: threading.Thread = None
    __executor = ThreadPoolExecutor(max_workers=1)
    __results = {}
    __callbacks = collections.defaultdict(list)
    __lock: threading.RLock = threading.RLock()

    __running = True

    __executions = {}

    # ------------------------------------------------------------------------
    def __init__(self, url: str, tag: str = None):
        self.__url = url
        self.__tag = url if tag == None else tag;
        self.__running = True
        self.__setup_call_backs()
        self.__sio.connect(self.__url, transports='websocket', namespaces=["/app"])
        self.__runner = threading.Thread(target=self.__run, args=())
        self.__runner.start()
        self.__checker = threading.Thread(target=self.__check_state, args=())
        self.__checker.start()

        # 3 seconds timeout
        maxRetries = 30
        while not self.is_connected() and maxRetries > 0:
            time.sleep(0.1)
            maxRetries -= 1

        if not self.is_connected():
            raise Exception(f"{self.__tag}: Failed to establish connection")

    def __del__(self):
        self.__lock.acquire()
        self.__executions.clear()
        self.__lock.release()
        self.__running = False
        self.__executor.shutdown(wait=False)
        self.__sio.disconnect()

    def __run(self): 
        logging.info(f"{self.__tag}: Proxy started")
        # Wait until the connection with the server ends.
        self.__sio.wait()
        logging.info(f"{self.__tag}: Proxy exiting")

    def __check_state(self):
        minimum_check_interval = 5       # seconds
        minimum_update_interval = 5      # seconds
        maximum_no_respose_interval = 15 # seconds
        last_update_check = datetime.now()

        logging.info(f"{self.__tag}: State checker started")
        while self.__running:
            #time.sleep(minimum_check_interval)
            while (datetime.now() - last_update_check).total_seconds() < minimum_check_interval and self.__running:
                time.sleep(0.1)

            last_update_check = datetime.now()

            executions_to_cancel = []
            executions_to_ping = []
            self.__lock.acquire()
            try:
                for rid, execution in self.__executions.items():
                    elapsedSinceLastUpdate = (datetime.now() - execution.__last_update).total_seconds()

                    # TODO: check reference count of the execution so see if it is still in use, and cancel it otherwise
                    # import sys
                    # if sys.getrefcount(execution) < ??:
                    #     executions_to_cancel.append(rid)
                    #     continue

                    if elapsedSinceLastUpdate > maximum_no_respose_interval:
                        executions_to_cancel.append(rid)
                    elif elapsedSinceLastUpdate > minimum_update_interval:
                        executions_to_ping.append(rid)
            except:
                logging.exception(f"{self.__tag}: Exception in state checker")
                pass
            self.__lock.release()

            try:
                for rid in executions_to_cancel:
                    logging.error(f"{self.__tag}: Execution {rid} is not responding. Cancelling")
                    self.__executions[rid].cancel()

                for rid in executions_to_ping:
                    qid = self.__executions[rid].request_qid()
                    if qid != None:
                        logging.info(f"{self.__tag}: Execution {rid} is not responding. Requesting update")
                        self.restore(qid)
                    else:
                        logging.info(f"{self.__tag}: Execution {rid} is not responding. Queue id not yet available")
            except:
                logging.exception(f"{self.__tag}: Exception in state checker")
                pass

        logging.info(f"{self.__tag}: State checker exiting")


    def request(self, input, uid):
        access=True
        if (uid == None):
            uid = uuid.uuid4().hex

        execution = ExecutionResult(self)
        rid = execution.request_id()

        # TODO: emit is not threadsafe. We need to lock here.
        #       need also to see if we catch exceptions here.
        data={}
        data["body"]=input
        data["header"]={"uid": uid, "rid": rid }
        data=zlib.compress(json.dumps(data).encode('utf-8')), access
        self.__sio.emit('execute', data=data, namespace='/app')

        # Add execution to monitored executions. Assuming rid is unique.
        self.__lock.acquire()
        self.__executions[rid] = execution
        self.__lock.release()

        return execution

    def cancel(self, execution : ExecutionResult):
        self.__lock.acquire()
        rid = execution.request_id()
        if rid in self.__executions:
            self.__executions.pop(rid)
        self.__lock.release()


    # TODO: remove
    def execute_async(self, input, uid = None, rid = None):
        access=True
        if (uid == None):
            uid = uuid.uuid4().hex
        if (rid == None):
            rid = uuid.uuid4().hex
        data={}
        data["body"]=input
        data["header"]={"uid": uid, "rid": rid }
        data=zlib.compress(json.dumps(data).encode('utf-8')), access
        self.__sio.emit('execute', data=data, namespace='/app')

        ++self.__pending_responses
        return self.__executor.submit(Proxy.get_response, self, rid)

    # TODO: remove
    def execute(self, input, uid = None, rid = None):
        return self.execute_async(input, uid=uid, rid=rid).result()

    def configure(self, config, uid):
        data={}
        data["body"]=config
        data["header"]={"uid": uid }
        data=zlib.compress(json.dumps(data).encode('utf-8'))
        self.__sio.emit('configure', data=data, namespace='/app')

    def is_connected(self):
        return self.__sio.connected

    def is_pending(self):
        return self.__pending_responses > 0

    def cancel_next(self):
        self.__cancel_next = True

    def get_response(self, rid: str):
        ++self.__pending_responses
        return self.__get_response(rid)


    # TODO: remove
    # TODO: should probably return also the error status
    def __get_response(self, rid: str):
        result=None
        print("Check state of rid: ", rid)
        while rid not in self.__results and not self.__cancel_next:
            # TODO: Given that we use a limited number of executors
            # per proxy, we should define a timeout.
            time.sleep(0.1)

        if self.__cancel_next:
            print("Canceled rid: ", rid)
        else:
            print ("Got response for rid: ", rid)
            result = self.__results.pop(rid)

        self.__cancel_next = False
        --self.__pending_responses

        return result

    def register(self, action, callback, context=None):
        self.__callbacks[action].append((callback, context))

    def unregister(self, action, callback):
        self.__callbacks[action] = [(f, c) for f, c in self.callbacks[action] if f != callback]

    def __notify(self, action, *args, **kwargs):
        for callback, context in self.__callbacks[action]:
            if context is not None:
                callback(context, *args, **kwargs)
            else:
                callback(*args, **kwargs)

    def resume(self, uid: str):
        self.__sio.emit('resume', data=uid, namespace='/app')

    def restore(self, qid: str):
        self.__sio.emit('restore', data=qid, namespace='/app')

    def state(self, uid: str):
        self.__sio.emit('state', data=uid, namespace='/app')

    def delete(self, qid: str):
        self.__sio.emit('delete', data=qid, namespace='/app')

    def __setup_call_backs(self):
        @self.__sio.event
        def connect():

            logging.info(f"{self.__tag}: Connection established")

        @self.__sio.event
        def connect_error(data):
            logging.info(f"{self.__tag}: Failed to establish connection")
            #raise Exception(f"{self.__tag}: Failed to establish connection")
            exit(0)

        @self.__sio.on("response", namespace='/app')
        def response(data):
            logging.debug(f"{self.__tag}: Data Received [response]: {data}")
            response: Dict[str, Any] = data
            # TODO: remove
            # TODO: Maybe we should include some other data and check also ray.status
            if "ray" in response and "output" in response:
                if "rid" in response["ray"]:
                    # TODO: would be good to store results only for known/recent requests
                    #       in order to avoid a case with faulty server sending bad data
                    self.__results[response["ray"]["rid"]] = (response["output"])

            rid=response["ray"]["rid"]

            self.__lock.acquire()
            if rid in self.__executions:
                self.__executions[rid].on_response(response)
            self.__lock.release()

            self.__notify("response", data)

        @self.__sio.on("submitted", namespace='/app')
        def submitted(data):
            logging.debug(f"{self.__tag}: Data Received [submitted]: {data}")

            response: Dict[str, Any] = data
            rid=response["rid"]

            self.__lock.acquire()
            if rid in self.__executions:
                self.__executions[rid].on_progress(response)
            self.__lock.release()

            self.__notify("submitted", data)

        @self.__sio.on("settings", namespace='/app')
        def settings(config):
            logging.debug(f"{self.__tag}: Data Received [settings]: {config}")
            response: Dict[str, Any] = config

            self.__notify("settings", response)

        # TODO: move functionality from sample app to here (without the remapping part)
        @self.__sio.on("progress", namespace='/app')
        def progress(data):
            logging.debug(f"{self.__tag}: Data Received [progress]: {data}")

            response: Dict[str, Any] = data
            rid=response["rid"]

            self.__lock.acquire()
            if rid in self.__executions:
                self.__executions[rid].on_progress(response)
            self.__lock.release()

            self.__notify("progress", data)

        @self.__sio.on("restore", namespace='/app')
        def restore(data):
            logging.debug(f"{self.__tag}: Data Received [restore]: {data}")
            response: Dict[str, Any] = data
            rid=response["ray"]["rid"]

            self.__lock.acquire()
            if rid in self.__executions:
                self.__executions[rid].on_restore(response)
            self.__lock.release()

            self.__notify("restore", data)

        @self.__sio.on("state", namespace='/app')
        def state(data):
            logging.debug(f"{self.__tag}: Data Received [state]: {data}")
            self.__notify("state", data)

        @self.__sio.event
        def disconnect():
            logging.info(f"{self.__tag}: Disconnected from server")


# ==============================================================================

def handler(data, *args, **kwargs):
    #logging.info(f"Handler called with data: {data}\n{args}\n------")
    pass

def handler2(data, *args, **kwargs):
    time.sleep(1000)

if __name__ == "__main__":
    # Enable logging
    logger = logging.getLogger()
    logger.setLevel(logging.INFO)

    # Sample proxy
    wrapper = Proxy('wss://0ff5d32a2fc74601b0781dfd2fc47974.openfabric.network', "App")

    '''
    wrapper.register("response", handler, "response")
    wrapper.register("progress", handler, "progress")
    wrapper.register("submitted", handler, "submitted")
    wrapper.register("restore", handler, "restore")
    wrapper.register("state", handler, "state")
    '''

    input={"text": ["string"]}

    logging.info("# execution = wrapper.request(input, uid=user)")
    execution = wrapper.request(input, uid="user")

    logging.info("---- > request_id " + str(execution.request_id()))
    logging.info("---- > progress " + str(execution.progress()))
    logging.info("---- > status " + str(execution.status()))
    logging.info("---- > result " + str(execution.data()))
    logging.info("---- > messages " + str(execution.messages()))

    logging.info("# execution.wait()")
    execution.wait()

    logging.info("---- > request_id " + str(execution.request_id()))
    logging.info("---- > progress " + str(execution.progress()))
    logging.info("---- > status " + str(execution.status()))
    logging.info("---- > result " + str(execution.data()))
    logging.info("---- > messages " + str(execution.messages()))

    logging.info("# execution = wrapper.request(input, uid=user)")
    execution = wrapper.request(input, uid="user")
    logging.info("# execution.cancel()")
    execution.cancel()

    logging.info("---- > request_id " + str(execution.request_id()))
    logging.info("---- > progress " + str(execution.progress()))
    logging.info("---- > status " + str(execution.status()))
    logging.info("---- > result " + str(execution.data()))
    logging.info("---- > messages " + str(execution.messages()))

    '''
    #wrapper.state("aaa")

    logging.info(f"Execute synced")
    response = wrapper.execute(input)
    logging.info(f"Response {response}")

    logging.info(f"Execute async")
    response = wrapper.execute_async(input)

    logging.info(f"Get async response")
    logging.info(f"Response {response.result()}")

    # Check exit is ok
    time.sleep(2)
    '''
    wrapper.__del__()




import logging
from typing import Any, Dict, Set

from openfabric_pysdk.flask.core import Webserver, request
from openfabric_pysdk.flask.socket import Namespace, SocketIOServer, SocketIOClient
from openfabric_pysdk.service import SocketService
from openfabric_pysdk.transport import ResourceDescriptor


#######################################################
#  Execution Socket
#######################################################
class ExecutionSocket(Namespace):
    __url: str = None
    __sessions: Set[str] = None
    __webserver: Webserver = None
    __sioserver: SocketIOServer = None
    __sioclient: SocketIOClient = None

    # ------------------------------------------------------------------------
    def __init__(self, webserver: Webserver, descriptor: ResourceDescriptor):
        super().__init__(descriptor.endpoint)
        self.__sessions = set()
        self.__webserver = webserver
        self.__url = f"{descriptor.remote}/{descriptor.endpoint}"

    # ------------------------------------------------------------------------
    def run(self, debug: bool, host: str, port: int):
        self.__sioserver = SocketService.server(self.__webserver)
        self.__sioserver.on_namespace(self)
        self.__sioclient = SocketService.client(self.__url, namespaces=[self.namespace])
        app_context = self.__webserver.app_context()

        @self.__sioclient.on("restore", namespace=self.namespace)
        def _restore(data):
            with app_context:
                self.__sioserver.emit('restore', data, namespace=self.namespace)

        @self.__sioclient.on("state", namespace=self.namespace)
        def _state(data):
            with app_context:
                self.__sioserver.emit('state', data, namespace=self.namespace)

        @self.__sioclient.on("progress", namespace=self.namespace)
        def _progress(data):
            with app_context:
                self.__sioserver.emit('progress', data, namespace=self.namespace)

        @self.__sioclient.on("response", namespace=self.namespace)
        def _response(data):
            with app_context:
                self.__sioserver.emit('response', data, namespace=self.namespace)

        @self.__sioclient.on("submitted", namespace=self.namespace)
        def _submitted(data):
            with app_context:
                self.__sioserver.emit('submitted', data, namespace=self.namespace)

        self.__sioserver.run(host=host, debug=debug, port=port)

    # ------------------------------------------------------------------------
    def on_execute(self, data: bytes, background: bool):
        self.__sioclient.emit('execute', data=(data, background), namespace=self.namespace)

    # ------------------------------------------------------------------------
    def on_resume(self, uid: str):
        self.__sioclient.emit('resume', data=uid, namespace=self.namespace)

    # ------------------------------------------------------------------------
    def on_restore(self, qid: str):
        self.__sioclient.emit('restore', data=qid, namespace=self.namespace)

    # ------------------------------------------------------------------------
    def on_state(self, uid: str):
        self.__sioclient.emit('state', data=uid, namespace=self.namespace)

    # ------------------------------------------------------------------------
    def on_delete(self, qid: str):
        self.__sioclient.emit('delete', data=qid, namespace=self.namespace)

    # ------------------------------------------------------------------------
    def on_connect(self):
        sid = request.sid
        logging.debug(f'Openfabric - client connected {sid} on {request.host}')
        self.__sessions.add(sid)

    # ------------------------------------------------------------------------
    def on_disconnect(self):
        sid = request.sid
        logging.debug(f'Openfabric - client disconnected {sid} on {request.host}')
        self.__sessions.remove(sid)



from typing import Any

from openfabric_pysdk.app import App
from openfabric_pysdk.context import StateStatus
from openfabric_pysdk.flask import Webserver
from openfabric_pysdk.flask import WebserverRestAPI, WebserverRestDoc
from openfabric_pysdk.loader import ConfigSchema
from openfabric_pysdk.service import ConfigService, RestService, SocketService, SwaggerService
from openfabric_pysdk.transport import ResourceDescriptor
from openfabric_pysdk.transport.socket.remote import ExecutionSocket as RemoteExecutionSocket
from openfabric_pysdk.transport.socket.local import ExecutionSocket as LocalExecutionSocket

from .profile import Profile


#######################################################
#  Container
#######################################################
class Container:
    __profile: Profile = None
    __websocket: [RemoteExecutionSocket, LocalExecutionSocket] = None
    __webserver: Webserver = None
    __rest: WebserverRestAPI = None
    __docs: WebserverRestDoc = None

    # ------------------------------------------------------------------------
    def __init__(self, profile: Profile, webserver: Webserver):
        self.__profile = profile
        self.__webserver = webserver
        self.__rest = WebserverRestAPI(webserver)
        self.__docs = WebserverRestDoc(webserver)

    # ------------------------------------------------------------------------
    def start(self, app: App):
        # Expose services
        self.__expose_swagger(app)
        self.__expose_rest(app)
        self.__expose_socket(app)
        # Apply configuration
        ConfigService.apply(app)

        app.set_status(StateStatus.RUNNING)

        # Start container
        # this will start eventing
        self.__websocket.run(
            debug=self.__profile.debug,
            host=self.__profile.host,
            port=self.__profile.port
        )

    # ------------------------------------------------------------------------
    def __expose_swagger(self, app: App):
        descriptor = self.__descriptor(app=app)
        SwaggerService.install(descriptor, webserver=self.__webserver)

    # ------------------------------------------------------------------------
    def __expose_socket(self, app: App):
        remote = self.__profile.remote
        if remote is None:
            from openfabric_pysdk.transport.socket.local import ExecutionSocket
        else:
            from openfabric_pysdk.transport.socket.remote import ExecutionSocket

        descriptor = self.__descriptor(handler=ExecutionSocket, endpoint='/app', app=app)
        self.__websocket = SocketService.install(descriptor, webserver=self.__webserver)

    # ------------------------------------------------------------------------
    def __expose_rest(self, app: App):

        remote = self.__profile.remote

        if remote is None:
            from openfabric_pysdk.transport.rest.local import \
                ConfigApi, \
                ExecutionApi, \
                ManifestApi, \
                BenchmarkApi, \
                QueueGetApi, QueuePostApi, QueueListApi, QueueDeleteApi
            from openfabric_pysdk.transport.socket.local import ExecutionSocket
        else:
            from openfabric_pysdk.transport.rest.remote import \
                ConfigApi, \
                ExecutionApi, \
                ManifestApi, \
                BenchmarkApi, \
                QueueGetApi, QueuePostApi, QueueListApi, QueueDeleteApi
            from openfabric_pysdk.transport.socket.remote import ExecutionSocket

        self.__install_rest(ExecutionApi, '/execution', app)
        self.__install_rest(ManifestApi, '/manifest', app)
        self.__install_rest(BenchmarkApi, '/benchmark', app)
        self.__install_rest(QueueGetApi, '/queue/get', app)
        self.__install_rest(QueueListApi, '/queue/list', app)
        self.__install_rest(QueuePostApi, '/queue/post', app)
        self.__install_rest(QueueDeleteApi, '/queue/delete', app)

        if ConfigSchema is not None:
            self.__install_rest(ConfigApi, '/config', app)

    # ------------------------------------------------------------------------
    def __install_rest(self, handler: type, endpoint: str, app: App):
        descriptor = self.__descriptor(handler=handler, endpoint=endpoint, app=app)
        RestService.install(descriptor, rest=self.__rest, docs=self.__docs)

    # ------------------------------------------------------------------------
    def __descriptor(self, handler=None, endpoint=None, app=None):
        descriptor = ResourceDescriptor()
        descriptor.app = app
        descriptor.handler = handler
        descriptor.endpoint = endpoint
        descriptor.remote = self.__profile.remote
        return descriptor



# Services
from flask_restful import Api as WebserverRestAPI
from flask_apispec import FlaskApiSpec as WebserverRestDoc

# Class definition
from flask_restful import Resource
from flask_apispec.views import MethodResource
from flask_apispec import doc, marshal_with, use_kwargs

# Swagger
from apispec import APISpec as WebserverSwaggerSpec
from apispec.ext.marshmallow import MarshmallowPlugin as SwaggerMarshmallowPlugin



import uuid

from openfabric_pysdk.engine import foreground
from openfabric_pysdk.flask.rest import *
from openfabric_pysdk.loader import *
from openfabric_pysdk.transport import ResourceDescriptor


#######################################################
#  Execution API
#######################################################
class ExecutionApi(MethodResource, Resource):
    __descriptor: ResourceDescriptor = None

    # ------------------------------------------------------------------------
    def __init__(self, descriptor: ResourceDescriptor = None):
        self.__descriptor = descriptor

    # ------------------------------------------------------------------------
    @doc(description="Execute execution and get response", tags=["Execution"])
    @use_kwargs(InputSchema, location='json')
    @marshal_with(OutputSchema)
    def post(self, *args) -> OutputClass:
        sid = uuid.uuid4().hex
        data = InputSchema().dump(list(args) if InputSchema().many is True else args[0])
        app = self.__descriptor.app
        qid = foreground.prepare(app, data, sid=sid)
        return foreground.process(qid)



from typing import Any, Dict, List, Union

from openfabric_pysdk.flask import *
from openfabric_pysdk.loader import *
from openfabric_pysdk.service import RemoteService
from openfabric_pysdk.transport import ResourceDescriptor
from openfabric_pysdk.transport.schema import UserId, UserIdSchema


#######################################################
#  Config API
#######################################################
class ConfigApi(MethodResource, Resource):
    __descriptor: ResourceDescriptor = None

    # ------------------------------------------------------------------------
    def __init__(self, descriptor: ResourceDescriptor = None):
        self.__descriptor = descriptor

    # ------------------------------------------------------------------------
    @doc(description="Get APP configuration", tags=["Developer"])
    @use_kwargs(UserIdSchema, location='query')
    @marshal_with(ConfigSchema)
    def get(self, uidc: UserId) -> ConfigClass:
        url = f"{self.__descriptor.remote}/{self.__descriptor.endpoint}?uid={uidc.uid}"
        return RemoteService.get(url, ConfigSchema().load)

    # ------------------------------------------------------------------------
    @doc(description="Set APP configuration", tags=["Developer"])
    @use_kwargs(UserIdSchema, location='query')
    @use_kwargs(ConfigSchema, location='json')
    @marshal_with(ConfigSchema)
    def post(self, uidc: UserId, config: Union[ConfigClass, List[ConfigClass]]) -> ConfigClass:

        if ConfigSchema().many is True and not isinstance(config, list):
            config = [config]
        else:
            config = config[0] if type(config) == list else config

        data = ConfigSchema().dump(config)
        url = f"{self.__descriptor.remote}/{self.__descriptor.endpoint}?uid={uidc.uid}"
        return RemoteService.post(url, data, ConfigSchema().load)
